/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

/**
 * @name fortisoar.appModulesService
 * @module fortisoar
 *
 * The `appModulesService` loads application modules and provides state information from module names.
 *
 **/
declare class appModulesService {
    /**
     * This function will load all the modules including system modules
     *
     * Returns a promise that is resolved with module metadata.
     * 
     * @param {Boolean=} excludeSystemModule System modules will be excluded if True.
     * @param {Boolean=} forceReload Based on the value modules will be fetched from localStorage or from database.
     **/
    static load(excludeSystemModule?: Boolean, forceReload?: Boolean): Promise<any>;

    /**
     * Get modules details from the local storage.
     *
     * Returns array of modules details.
     * 
     * @param {Array} modules Array of module name.
     **/
    static getModules(modules: Array<String>): Array<Object>;

    /**
     * Returns state of the provided module name.
     * 
     * @param {String} moduleName The name for a module.
     **/
    static getState(moduleName: String): String;

    /**
     * Return the list view state based on the given module name.
     *
     * @param {string} moduleName The name for a module.
     **/
    static getListState(moduleName: String): String;

    /**
     * This function stores the formatted error message based on the given error object in the local storage.
     *
     * @param {Object} rejection Object details of the error.
     * @param {Array=} attributes Array of fields.
     * @param {Boolean=} isConfigurationManager Set the publish error details in the local storage.
     **/
    static checkViolations(rejection: Object, attributes?: Array<Object>, isConfigurationManager?: Boolean): void;

    /**
     * Returns a promise that is resolved with queue module metadata.
     **/
    static getQueueModules(): Promise<any>;
}

/**
 * @name fortisoar.CommonUtils
 * @module fortisoar
 *
 * The `commonUtils` service contains commonly used functions.
 **/
declare class CommonUtils {
    /**
     * This function check the provided input is valid UUID or not.
     * 
     * Returns true if provided input is valid UUID, else false.
     * 
     * @param {String} s It contains UUID value to check.
     *
     **/
    static isUUID(s: String): Boolean;

    /**
     * This function create and return UUID.
     *
     **/
    static generateUUID(): String;

    /**
     * Returns true/false based on the provided value is undefined or null.
     *
     * @param {*} value Value to check for undefined.
     * 
     **/
    static isUndefined(value: any): Boolean;

    /**
     * Returns true/false based on the provided value is number or not.
     *
     * @param {*} value Value to check for number or not.
     * 
     **/
    static isNumber(value: any): Boolean;

    /**
     * Convert and returns the JSON object.
     *
     * @param {String} str String to convert into JSON format.
     * 
     **/
    static parseJSON(str: String): Object;

    /**
     * This function is used for copy to clipboard.
     *
     * @param {Object} fields It contains fields value.
     * @param {Object=} row It contains rows.
     * @param {Object=} col It contains columns.
     * @param {Object=} rows Array of row objects.
     * 
     **/
    static copyToClipboard(fields: Object, row?: Object, col?: Object, rows?: Object): void;

    /**
     * Returns boolean value based on the provided value is in jinja format or not.
     * 
     * @param {String} value Input value to check for Jinja format.
     *
     **/
    static isJinja(value: String): Boolean;

    /**
     * Returns boolean value based on the provided value is in object format or not.
     * 
     * @param {Object} value Input value for Object validation.
     *
     **/
    static isObject(value: Object): Boolean;

    /**
     * Returns boolean value based on the provided value is valid URL or not.
     * 
     * @param {String} url Input value to validate URL.
     *
     **/
    static isValidURL(url: String): Boolean;

    /**
     * Returns boolean value based on the provided value is in Base64 image format or not.
     * 
     * @param {String} imageString Input value to validate base64 format.
     *
     **/
    static isBase64Image(imageString: String): Boolean;

    /**
     * Returns an object with a key as a searchText and value from the provided JSON.
     * 
     * @param {Object} jsonObj JSON object.
     * @param {String} searchText Key value for search in the object.
     *
     **/
    static searchInJSON(jsonObj: Object, searchText: String): Object;
}

/**
 * @name fortisoar.connectorService
 * @module fortisoar
 *
 * The purpose of the service is to handle connector-specific operations like creating connectors, getting connector details, etc.
 *
 **/
declare class connectorService {
    /**
     * Returns a promise that will be resolved with the connector details.
     * 
     * @param {String} name This is the name of the connector. Based on this name the connector 
     * details are fetched.
     * @param {String} version This is the version of the connector.
     * @param {String=} agent This is the agent ID. If provided, based on this ID the connector 
     * details are fetched from the agent machine.
     *
     **/
    static getConnector(name: String, version: String, agent?: String): Promise<any>;

    /**
     * Returns a promise that will be resolved with the connector details.
     *
     * @param {String} id This is the ID of the connector. Based on this ID the connector details are fetched.
     * @param {Boolean=} editRepoConnector Flag to mention regarding edit action of connector.
     * 
     **/
    static getDevelopedConnector(id: String, editRepoConnector?: Boolean): Promise<any>;

    /**
     * Returns a promise that will be resolved with the connector's health details.
     *
     * @param {Object} connector This is the connector object.
     * @param {String} configUUID Configuration ID of the connector.
     * @param {String=} agent This is the agent ID. If provided, based on this ID 
     * the connector's health details are fetched from the agent machine.
     * 
     **/
    static getConnectorHealth(connector: Object, configUUID: String, agent?: String): Promise<any>;

    /**
     * Returns a promise that will be resolved on the connector update success, else reject.
     *
     * @param {Object} connector This is the connector object.
     * @param {Boolean} newConfig Indicates whether the configuration is new or already exists.
     * @param {Boolean} deleteConfig Set true to delete the selected configuration.
     * 
     **/
    static updateConnectorConfig(connector: Object, newConfig: Boolean, deleteConfig: Boolean): Promise<any>;

    /**
     * Returns a promise that will be resolved with the executed operation details.
     *
     * @param {String} connectorName Name of the connector.
     * @param {String} connectorVersion Version of the connector.
     * @param {String} operation Operation name.
     * @param {String} configuration Configuration ID.
     * @param {Object} params Parameters for the operation.
     * @param {Boolean=} audit This will check weather this will audited or not.
     * @param {Object=} auditInfo Audit information.
     * @param {String=} agent This is the agent ID. Based on this ID the connector details are fetched from the agent machine.
     * 
     **/
    static executeConnectorAction(connectorName: String, connectorVersion: String, operation: String, configuration: String, params: Object, audit?: Boolean, auditInfo?: Object, agent?: String): Promise<any>;

    /**
     * Returns a promise that will be resolved with the status of delete operation.
     *
     * @param {String} id This is the ID of the connector.
     * 
     **/
    static deleteConnector(id: String): Promise<any>;
}

/**
 * @name fortisoar.currentPermissionsService
 * @module fortisoar
 *
 * The purpose of the `currentPermissionsService` is to get module specific permission details.
 *
 **/
declare class currentPermissionsService {
    /**
     * Returns details of modules with it's permission.
     *
     **/
    static get(): Object;

    /**
     * Returns details of module with it's permission.
     * 
     * @param {String} module Name of the module to fetch permissions.
     *
     **/
    static getPermission(module: String): Object;

    /**
     * Returns details of modules with it's permissions.
     *
     * @param modules Array of module names to fetch permissions.
     * 
     **/
    static getPermissions(modules: Array<String>): Object;

    /**
     * The purpose of this method is to find the permission for specific actions 
     * like 'read/write/update/delete' for the provided module.
     * 
     * Returns true if module have specified permission else return false.
     * 
     * @param {String} module Module name.
     * @param {String} action Type of permission e.g read, delete, update, create
     *
     **/
    static availablePermission(module: String, action: String): Boolean;

    /**
     * The purpose if method is to get the status of logged in user's' security update permission.
     * 
     * Returns true if logged in user have security update permission else return false.
     *
     **/
    static isAdmin(): Boolean;

    /**
     * The purpose of this method is to find the permission for specific actions 
     * like 'read/write/update/delete' for the provided list of modules.
     * 
     * Returns list of modules which have the provided permission.
     *
     * @param {Array} modules Array of all modules.
     * @param {String} action Type of permission e.g read, delete, update, create
     * 
     **/
    static availablePermissions(modules: Array<Object>, action: String): Array<Object>;

    /**
     * The purpose of this method is to find the permission for specific actions 
     * like 'read/write/update/delete' for the provided modules field.
     * 
     * Returns true if field have provided permission else return false.
     *
     * @param {String} module Name of the module.
     * @param {String} fieldName Field name of the module.
     * @param {String} action Type of permission e.g read, delete, update, create
     * 
     **/
    static availableFieldPermission(module: String, fieldName: String, action: String): Boolean;
}

/**
 * @name fortisoar.editorLanguageService
 * @module fortisoar
 *
 * The `editorLanguageService` handles the language server functionality to communicate with the 
 * language server for providing IntelliSense in the Editor.
 *
 **/
declare class editorLanguageService {
    /**
     * The method initialises Monaco Editor services.
     * 
     **/
    static initMonacoEditorService(): void;

    /**
     * The method handles connection with language server. Creates new socket if its not already created.
     * Or connects with language server if socket is already present but not connected.
     * 
     **/
    static connectLangSocket(): void;

    /**
     * The method handles editor closure count.
     * 
     **/
    static closeEditor(): void;

    /**
     * The method handles websocket reconnection and listen for events and message communication.
     * 
     **/
    static createAndListenOnLangSocket(): void;

    /**
     * The method closes websocket connection with language service on successfully closing all editor connections.
     * 
     **/
    static closeLangSocket(): void;
}

/**
 * @name fortisoar.Entity
 * @module fortisoar
 *
 * The `Entity` service is a prototyped object that represents a module and all
 * of its metadata and fields. Handles loading and population of data
 * based on permissions, module name, and id. Takes care of adding, and removing
 * relationships, and transforming form data.
 *
 **/
declare class Entity {
    /**
     * Returns promise object that on success will populate the fields properties on the entity instance based on the loaded data.
     *
     * @param {String} id UUID of the record.
     * @param {Object=} userParams A filter object to add extra filters such as `{ $relatioships: true }`.
     * @param {Boolean=} ignoreErrors If true, errors will be ignored while fetching module data, else errors will be catched.
     * @param {Boolean=} refresh If true, latest fields will be loaded.
     **/
    static get(id: String, userParams?: Object, ignoreErrors?: Boolean, refresh?: Boolean): Promise<any>;

    /**
     * Returns resolved promise in case fields are loaded successfully, else returns reject promise.
     *
     * @param {String=} mode Mode of operation like 'add', 'edit', etc.
     * @param {Boolean=} ignoreRBAC If true, field permissions will be ignored.
     **/
    static loadFields(mode?: String,ignoreRBAC?: Boolean): Promise<any>;

    /**
     * Returns resolved promise in case relationships fields are loaded successfully, else returns reject promise.
     **/
    static loadAllRelationships(): Promise<any>;

    /**
     * Returns resolved promise on success with fields relationship data, else returns reject promise.
     *
     * @param {String} fieldName Name of the field.
     * @param {Boolean=} getObjects Get result in object format.
     * @param {Number=} recordSize Maximum record size to fetch.
     * @param {Array=} columns Array of field name only which we want to fetch and ignore other fields.
     **/
    static getRelationship(fieldName: String, getObjects?: Boolean, recordSize?: Number, columns?: Array<String>): Promise<any>;

    /**
     * Returns resolved promise on success, else returns reject promise.
     *
     * @param {Object} data Name of the field.
     * @param {Object=} params Other optional parameters.
     * @param {Array=} fieldsToKeep List of fields which we don't want to update.
     **/
    static save(data: Object, params?: Object, fieldsToKeep?: Array<String>): Promise<any>;

    /**
     * Promise object that on success will delete the entity instance.
     **/
    static delete(): Promise<any>;

    /**
     * Returns object of form fields.
     * 
     * @param {Boolean=} sorted To get the result in sorted format.
     *
     **/
    static getFormFields(sorted?: Boolean): Object;

    /**
     * Returns object of form fields without system fields.
     **/
    static getFormFieldsWithoutSystem(): Object;

    /**
     * Returns object of relationship fields.
     **/
    static getRelationshipFields(): Object;

    /**
     * Returns array of form fields.
     * 
     * @param {String=} type Entity type.
     * 
     **/
    static getFormFieldsArray(type?: String): Array<String>;

    /**
     * Returns array of form fields without system fields.
     * 
     * @param {String=} type Entity type.
     * 
     **/
    static getFormFieldsArrayWithoutSystem(type?: String): Array<String>;

    /**
     * This function will set the default value to all the fields.
     **/
    static applyDefaultValues(): void;

    /**
     * Returns array of relationship field objects.
     **/
    static getRelationshipFieldsArray(): Array<Object>;

    /**
     * Returns promise object that on success will export in CSV fortmat.
     **/
    static getExportData(): Promise<any>;

    /**
     * Returns the array of all the fields title.
     **/
    static getFieldTitles(): Array<String>;

    /**
     * Returns object of the evaluated query filters.
     * 
     * @param {Object} query Object of query filters.
     *
     **/
    static evaluate(query: Object): Object;

    /**
     * This function will evaluate the visibility and required filters for all the entity fields.
     * 
     **/
    static evaluateAllFields(): void;

    /**
     * Returns boolean value based on whether is there any required field with no value present.
     **/
    static hasEmptyRequiredFields(): Boolean;

    /**
     * Returns array of fields which are required with no value.
     **/
    static getEmptyRequiredFields(): Array<String>;

    /**
     * Returns object of fields which are required with no value
     **/
    static getEmptyRequiredFieldsObject(): Object;
}

/**
 * @name fortisoar.exportService
 *
 * The `exportService` handles copying, exporting, and downloading playbooks
 *
 **/
declare class exportService {
    /**
     * Copies 'entities' of type 'entityName' (workflows) and returns the result.
     * 'workflows' require the associated steps and routes to be copied, and 
     * workflows need their UUIDs updated.
     *
     * Returns array of copied entity objects.
     * 
     * @param {Array} entities An array of entity objects to copy.
     * @param {String} entityName The name of the entity (workflow).
     * @param {Boolean=} changeName Indicates whether clone or not. If set to true, will change the playbook name to make it unique.
     * @param {Object=} uuidPlaybookMap Playbook UUID map.
     **/
    static copyEntities(entities: Array<Object>, entityName: String, changeName?: Boolean, uuidPlaybookMap?: Object): Array<Object>;

    /**
     * Returns promise object that on success will populate the collection names.
     *
     **/
    static loadCollectionNames(): Promise<any>;

    /**
     * Returns promise which provides a list of playbooks for the provided collection on resolve.
     *
     * @param {Object} collection Object which contains details of the collection like name, uuid, etc.
     * 
     **/
    static loadCollectionPlaybooks(collection: Object): Promise<any>;

    /**
     * Change the existing playbook UUID with new UUID while exporting the playbook.
     *
     * @param {Object} playbook Playbook object to export.
     * @param {Object} uuidPlaybookMap Playbook UUID map.
     * @param {Boolean} replaceAllUuids If true, then replace the existing UUID with new UUID.
     * @param {Boolean} collectionClone If true, then delete playbook version based on the input.
     * @param {Boolean} existingOutsideCollection If true, then replace the existing UUID with new UUID.
     * 
     **/
    static changePlaybookAndStepsUuid(playbook: Object, uuidPlaybookMap: Object, replaceAllUuids: Boolean, collectionClone: Boolean, existingOutsideCollection: Boolean): void;

    /**
     * Returns promise that will resolve with the exported hydra response.
     *
     * @param {Array} rows An array of entity objects to export. Each object must have '@id' key.
     * @param {String} entityName The name of the entity (workflow).
     * @param {Array=} selectedFields Array of entity field names.
     * @param {Boolean=} skipExportFlag To set export flag in query.
     * @param {Boolean=} includeVersions Include versions of the playbook while exporting.
     * @param {Boolean=} childEntityName Include relationships of the playbook while exporting.
     **/
    static loadRowsForExport(rows: Array<Object>, entityName: String, selectedFields?: Array<String>, skipExportFlag?: Boolean, includeVersions?: Boolean, childEntityName?: Boolean): Promise<any>;

    /**
     * Returns promise that will resolve with the exported hydra response
     *
     * @param {Object} playbooks Playbook object to export.
     * @param {Boolean=} includeVersions Include versions of the playbook while exporting.
     * @param {Array=} previousIds Array of previous step Uuids.
     **/
    static getReferencePlaybookForExport(playbooks: Object, includeVersions?: Boolean, previousIds?: Array<String>): Promise<any>;

    /**
     * Forces download of a JSON file with given data object and fileName.
     *
     * @param {Object} data The object that will be converted to JSON.
     * @param {String} fileName The name that the file will be saved with.
     **/

    static downloadJsonFile(data: Object, fileName: String): void;

    /**
     * Returns list of macros from the playbook steps.
     *
     * @param {Object} playbook Playbook object with all steps details.
     * @param {Array} variables List of variables to compare with macros.
     **/
    static getMacrosFromPlaybook(playbook: Object, variables: Array<Object>): Array<Object>;

    /**
     * Returns promise that will resolve with successfully save macros.
     *
     * @param {Array} macros Array of macros.
     **/
    static saveNewMacrosFromPlaybook(macros: Array<Object>): Promise<any>;

    /**
     * Returns promise that will resolve with successfully export the record.
     *
     * @param {String} route Action trigger route.
     * @param {Object} params Playbook collection details and record details to export.
     **/
    static exportGridRecords(route: String, params: Object): Promise<any>;
}

/**
 * @name fortisoar.Field
 * @module fortisoar
 *
 * The `Field` service is an object that represents the model of a field.
 *
 **/
declare class Field {
    /**
     * Returns the raw value of the Field. This could
     * be a datetime timestamp, an IRI, a plain string, or even the entire object.
     **/
    static getFormValue(): any;

    /**
     * This function will evaluate the required constraints of the field.
     * 
     * @param {Object} entity An object of entity.
     *
     **/
    static evaluateRequired(entity: Object): void;

    /**
     * This function will evaluate the visible constraints of the field.
     * 
     * @param {Object} entity An object of an entity.
     *
     **/
    static evaluateVisible(entity: Object): void;
}

/**
 * @name fortisoar.FormEntityService
 * @module fortisoar
 *
 * The `FormEntityService` handles entity related operation like set, get and submit fields.
 *
 **/
declare class FormEntityService {
    /**
     * This function saves the provided entity.
     * 
     * @param {Object} newEntity Entity object.
     *
     **/
    static set(newEntity: Object): void;

    /**
     * The purpose of this method is to get entity.
     * 
     * Returns entity object.
     *
     **/
    static get(): Object;
}

/**
 * @name fortisoar.licenseService
 * @module fortisoar
 *
 * The `licenseService` gets and sets license details.
 *
 **/
declare class licenseService {
    /**
     * Returns promise with license data.
     *
     * @param {Boolean=} [reload=undefined] To reload license details.
     * @param {Boolean=} [includeNodes=undefined] To get nodes license details.
     **/
    static getLicenseDetails(reload?: Boolean, includeNodes?: Boolean): Promise<any>;

    /**
     * Returns promise with branding details of the instance.
     * 
     * @param {Boolean=} [latest=undefined] For fetching latest branding details.
     */
    static getBrandingDetails(latest?: Boolean): Promise<any>;
}

/**
 * @name fortisoar.modelMetadatasService
 * @module fortisoar
 *
 * The `modelMetadatasService` loads model metadata and retrieves information from it.
 *
 **/
declare class modelMetadatasService {
    /**
     * Checks local storage for the model metadata. If they are not in
     * storage, it falls back to checking whether another promise to get them
     * has already been set. If one has not been set, then an API request is
     * made to retrieve them and store them in local storage.
     *
     * Returns a promise that is resolved with model metadata.
     * 
     * @param {Boolean=} staging True for staging environment.
     * @param {Boolean=} forceUpdateModelMeta If true, then data fetched from server API call, else used from local storage.
     * @param {Boolean=} mmdEditorCall Handle MMD Editor call, to restrict local stprage update.
     **/
    static getModuleList(staging?: Boolean, forceUpdateModelMeta?: Boolean, mmdEditorCall?: Boolean): Promise<any>;

    /**
     * Loads all modules along with the system modules.
     * 
     * Returns a promise that is resolved with module metadata.
     * 
     * @param {Boolean=} forceUpdate On true fetch latest model metadata else uses data from localstorage.
     **/
    static loadAllModules(forceUpdate?: Boolean): Promise<any>;

    /**
     * Retrieves a model metadata type from a given module type.
     *
     * Returns module metadata as per the provided module type.
     * 
     * @param {String} type The type of the model metadata by which to retrieve the type.
     **/
    static getMetadataByModuleType(type: String): Object;

    /**
     * Retrieves a model metadata IRI from a given module type.
     *
     * Returns the IRI of the model metadata.
     * 
     * @param {String} type The type of the model metadata by which to retrieve the IRI.
     **/
    static getIriByType(type: String): String;

    /**
     * Return a promise that is resolved with the module metadata of the tenant.
     *
     * @param {String=} tenantUuid This is the uuid of the tenant.
     **/
    static getTenantModuleList(tenantUuid: String): Promise<any>;

    /**
     * This function will start the publish on tenant based on the tenant UUID.
     *
     * Returns a promise that is resolved with success of publish.
     * 
     * @param {String} tenantUuid This is the uuid of the tenant.
     **/
    static publishTenant(tenantUuid: String): Promise<any>;

    /**
     * This function will return the stagging module metadata of the tenant based tenant uuid and module id.
     *
     * Returns a promise that is resolved with success will populate the stagging module metadata.
     * 
     * @param {String} tenantUuid This is the uuid of the tenant.
     * @param {String} moduleUuid This is the uuid of the module.
     **/
    static getTenantStaggingModule(tenantUuid: String, moduleUuid: String): Promise<any>;

    /**
     * Retrieves a model name in singular or plural form from local storage by using a given module type.
     *
     * @param {Array} moduleTypes Array of the modules
     * @param {Boolean} isPlural boolean value to return name in singular or plural form
     * 
     * Returns the array of module data object contains type and name as per isPlural flag.
     **/
    static getModuleNameByType(moduleTypes: Array<String>, isPlural: Boolean): Array<Object>;
}

/**
 * @name fortisoar.Modules
 * @module fortisoar
 *
 * The `Modules` service is simply a wrapper for ng.$resource and adds update.
 * Optionally accepts `module`, `id`, and `fieldName`
 **/
declare class Modules {
    /**
     * @ngdoc method
     * @name fortisoar.Modules#save
     * @methodOf fortisoar.Modules
     *
     * @description
     * This function will update/save the module.
     *
     * @param {Object} config This object will contain the module name as `module: module_name`.
     * The object can also contains `$relationships: true` property.
     * @param {Object} params This object will contain data to be save or update.
     * @returns {Promise} Promise object that on success will update/save the module.
     **/
    static save(config: Object, params: Object): Promise<any>;
}

/**
 * @name fortisoar.picklistsService
 * @module fortisoar
 *
 * The `picklistsService` will facilitate performing actions in regards gathering
 * and manipulating picklists.
 **/
declare class picklistsService {
    /**
     * Returns a promise that will be resolved with the picklist once found,
     * or rejected if not found.
     *
     * @param {Field} field The field object that has picklists.
     * @param {String} itemValue The itemValue property of the picklist to be found.
     **/
    static getPicklistByItemValue(field: Field, itemValue:String): Promise<any>;

    /**
     * Returns a promise that will be resolved with the picklist object
     * once found, or rejected if not found.
     *
     * @param  {String} iri The IRI of the picklist to find. For example: `/api/3/picklist_names/c2e3bcad-778d-41dd-91e8-0596b8c5716f`
     **/
    static getPicklistByIri(iri: String): Promise<any>;

    /**
     * Returns a promise that will be resolved with the picklist values.
     *
     * @param  {Boolean=} useExport $export flag to be passed as a part of query parameters while fetching picklist data.
     **/
    static loadAllPicklists(useExport?: Boolean): Promise<any>;

    /**
     * Returns a promise that will be resolved with the picklist object
     * once found, or rejected if not found.
     *
     * @param  {String} itemVal The IRI of the picklist to find. For example: `/api/3/picklist_names/c2e3bcad-778d-41dd-91e8-0596b8c5716f`
     * @param  {String} itemKey The key of the picklist to find. For example: `name`
     **/
    static loadPicklistsByParams(itemVal: String, itemKey: String): Promise<any>;
}

/**
 * @name fortisoar.playbookService
 * @module fortisoar
 *
 * The purpose of the `playbookService` is to perform playbook related operation like get playbook details, create playbook, get playbook execution log details etc.
 *
 **/
declare class playbookService {
    /**
     * Returns resolved promise with the playbooks details on success.
     *
     * @param {Object} query Query parameters to fetch the running playbooks details.
     * 
     **/
    static getRunningPlaybooks(query: Object): Promise<any>;

    /**
     * Returns a promise that will be resolved with the playbooks execution count details.
     *
     **/
    static getPlaybookExecutionCount(): Promise<any>;

    /**
     * Returns a promise that will be resolved with the step details of the playbooks.
     *
     * @param {String} id Step ID of the playbook.
     * 
     **/
    static getStepRunningDetails(id: String): Promise<any>;

    /**
     * @param {String} id ID of the playbook.
     * 
     * Returns a promise that will be resolved with the playbook details.
     *
     **/
    static getRunningPlaybookDetails(id: String): Promise<any>;
}

/**
 * @name fortisoar.PromiseQueue
 * @module fortisoar
 *
 * The `PromiseQueue` service stores, retrieves, and clears promises so they
 * can be shared instead of doing the same work multiple times. This is very
 * useful when making API calls in quick succession when the data being returned
 * is not expected to be different.
 *
 **/
declare class PromiseQueue {
    /**
     * Returns a promise that has previously been stored using a key.
     * Will return undefined if an invalid key is used.
     *
     * @param {String} key The unique key that was used to store the promise.
     **/
    static get(key: String): Promise<any>;

    /**
     * Sets a promise into the queue for later retrieval.
     *
     * @param {String} key The unique key to use to store the promise.
     * @param {Promise} promise The promise to be stored.
     **/
    static set(key: String, promise: Promise<any>): void;

    /**
     * Reinitializes the queue.
     * 
     * @param {String} key The unique key that was used to store the promise.
     **/
    static clear(key: String): void;
}

/**
 * @name fortisoar.Query
 * @module fortisoar
 *
 * The `Query` service provides an object that represents API query parameters
 * and filters and exposes methods for manipulating and retrieving data about
 * them.
 **/
declare class Query {
    /**
     * Updates filter array.
     * 
     * @param {Array} filters Requires updated array of filter objects.
     */
    static updateFilters(filters: Array<Object>): void;

    /**
     * Returns updated filter object.
     * 
     * @param {Object} filter Provide a filter condition.
     * @param {String} type Provide field type.
     */
    static updateFilter(filter: Object, type: String): Object;

    /**
     * Returns proper evaluated query.
     * 
     * @param {Object=} [entity=undefined] Takes input as `entity` Object.
     */
    static getFlatQuery(entity?: Object): Object;

    /**
     * Modifies query modifiers i.e., limit to $limit, page to $page.
     * Returns `query` object with modifiers.
     * 
     * @param {Object=} [entity=undefined] Takes input as `entity` object.
     */
    static getQueryModifiers(entity?: Object): Object;

    /**
     * Forms query object with all required parameters.
     * Query Object Parameters: 
     * sort, limit, logic, search, showDeleted, filters, $relationships, aggregates, __selectFields
     * Return query object.
     * 
     * @param {Boolean=} useValue Flag used for datetime identifier in the conditional filter.
     * @param {String=} dataMode Takes input as string i.e., 'edit', 'view' etc.
     */
    static getQuery(useValue?: Boolean, dataMode?: String): Object;

    /**
     * Update query param object's sort data as per required order for POST request call.
     * 
     * @param {Object} entity Requires `entity` object.
     * @param {Object} queryParam Requires query object.
     */
    static buildOrderForPostRequest(entity: Object, queryParam: Object): void;
}

/**
 * @name fortisoar.queryCollectionService
 * @module fortisoar
 *
 * The `queryCollectionService` loads, saves, manipulates, and deletes query collections, etc..
 *
 **/
declare class queryCollectionService {
    /**
     * This will load the queries for a given module and store a promise for all
     * subsequent requests.
     * Returns a promise that will be resolved with the queries once loaded.
     *
     * @param {String} model The name of the model from which queries are to be loaded.
     **/
    static load(model: String): Promise<any>;

    /**
     * Updates the values of the provided query's filters with either a picklist
     * object, or display value.
     * Returns a promise that will be resolved after updating all
     * filter values.
     *
     * @param {Query} query The query object to update filter values
     **/
    static loadQueryFilterValues(query: Query): Promise<any>;
}

/**
 * @name fortisoar.services:PagedCollection
 * @module fortisoar
 *
 * PagedCollection service creates objects that work against the paged crud-hub API.
 * This is the base class that other PagedCollection child classes would use.
 *
 **/
declare class PagedCollection {
    /**
     * Returns properly ordered grid page promise.
     * 
     * @param {Object=} config Takes input as grid configuration.
     * @param {Boolean=} sortApplied Takes input to apply sort to record.
     * @param {Boolean=} getAllColumns Takes input to load all columns.
     * @param {Number=} recordId Takes input to load specific record with `recordId`.
     * @param {String=} associationField Takes input to load specific field's association.
     * @param {Boolean=} [csvNeutralization=undefined] Takes input to load neutralized CSV data.
     **/
    static loadGridRecord(config?: Object, sortApplied?: Boolean, getAllColumns?: Boolean, recordId?: Number, associationField?: String, csvNeutralization?: Boolean): Promise<any>;

    /**
     * Loads grid data by `POST` call.
     * Returns properly ordered grid page promise.
     * 
     * @param {Object} config Takes input as query configuration.
     * @param {Boolean=} [sortApplied=undefined] Takes input to sort data.
     **/
    static loadByPost(config: Object, sortApplied?: Boolean): Promise<any>;

    /**
     * Loads all records with customQuery, customColumns and relationship.
     * Returns updated pagedCollection records.
     * 
     * @param {Object=} customQuery Takes input as query params.
     * @param {Array=} customColumns Takes input as columns as fieldName.
     * @param {Boolean=} relationshipKey Takes input to load relationship.
     **/
    static load(customQuery?: Object, customColumns?: Array<String>, relationshipKey?: Boolean): Promise<any>;

    /**
     * Returns record as per pagination.
     * 
     * @param {Object} data Takes input as API result as data.
     **/
    static setPage(data: Object): Promise<any>;

    /**
     * Navigate to data of first page.
     * Returns first page data as promise.
     * 
     * @param {Boolean} loadByPost Calls loadGridRecord method on true, else forms first page URL and calls API to get results.
     **/
    static pageFirst(loadByPost: Boolean): Promise<any>;

    /**
     * Navigate to data of previous page.
     * Returns previous page as promise.
     * 
     * @param {Boolean} loadByPost Calls loadGridRecord method on true, else forms previous page URL and calls API to get results.
     **/
    static pagePrevious(loadByPost: Boolean): Promise<any>;

    /**
     * Navigate to data of next page.
     * Returns next page as promise.
     * 
     * @param {Boolean} loadByPost Calls loadGridRecord method on true, else forms next page URL and calls API to get results.
     **/
    static pageNext(loadByPost: Boolean): Promise<any>;

    /**
     * Navigate to data of last page.
     * Returns last page data as promise.
     * 
     * @param {Boolean} loadByPost Calls loadGridRecord method on true, else forms last page URL and calls API to get results.
     **/
    static pageLast(loadByPost: Boolean): Promise<any>;

    /**
     * Navigate to specific page for data.
     * Returns specific page data as promise.
     * 
     * @param {Object} params Calls loadGridRecord method on objects 'loadByPost' property, 
     * else forms page URL as per provided properties and calls API to get results.
     **/
    static gotoPage(params: Object): Promise<any>;

    /**
     * Converts `hydra:member` records to key pairs.
     * 
     * Returns promise for converting hydra:member to Field rows.
     **/
    static convertToKeyPairs(): void;

    /**
     * Builds sort query.
     * 
     * @param {Array} sortColumns Takes input as columns.
     **/
    static buildSortQuery(sortColumns: Array<Object>): void;

    /**
     * Sorts column by `field name`.
     * 
     * @param {String} direction Takes input as sorting direction i.e., 'desc'.
     **/
    static sortColumnsByFieldName(direction: String): void;

    /**
     * Loads default columns.
     * Returns updated object of default columns.
     * 
     * @param {String=} relationshipKey Takes input of related field name key.
     * @param {Boolean=} mergeDefaults Input takes a boolean whether to merge defaults with existing.
     **/
    static loadDefaultColumns(relationshipKey?: String, mergeDefaults?: Boolean): Object;

    /**
     * Extends filter condition.
     * 
     * @param {Object} _config Takes input as `query` object.
     * @param {*} value Add `null` as value, not required, doesn't have any impact.
     * @param {Object} field Takes input as `field` object.
     **/
    static extendFilter(_config: Object, value: any, field: Object): void;
}

/**
 * @name fortisoar.settingsService
 * @module fortisoar
 *
 * The `settingsService` gets and sets user and system settings.
 *
 **/
declare class settingsService {
    /**
     * Returns the value of the setting or empty string if not found.
     *
     * @param {String} key string representing a key in a user's settings.
     * The key is delimited by `/`. Given the example below,
     * to get the theme value you would provide `user/view/theme` as the key.
     * @example
     * user: {
     *   view: {
     *     theme: 'dark'
     *   }
     * }
     **/
    static get(key: String): String;

    /**
     * Checks local storage for the system settings. If they are not in
     * storage, it falls back to checking whether another promise to get them
     * has already been set. If one has not been set, then an API request is
     * made to retrieve them and store them in local storage.
     *
     * Returns a promise that is resolved with the system settings.
     * 
     * @param {Boolean=} forceRefresh If true then data fetched from server, else uses browsers local storage data.
     **/
    static getSystem(forceRefresh?: Boolean): Promise<any>;
}

/**
 * @name fortisoar.tokenService
 * @module fortisoar
 *
 * The `tokenService` handles idle and session timers, and provides methods for
 * manipulating and getting information from JWT web tokens.
 *
 **/
declare class tokenService {
    /**
     * The method sets the encrypted token and user information in local storage. 
     * Also, create timers to validate tokens for refresh and adding listeners.
     * 
     * @param {String} token Contains plain token received in API.
     *
     **/

    static set(token: String): void;

    /**
     * The purpose of this method is to get authentication token.
     * 
     * Returns plain token.
     *
     **/
    static get(): String;
}

/**
 * @name fortisoar.usersService
 * @module fortisoar
 *
 * The `usersService` is an auxiliary service that performs actions regarding
 * users. It includes methods for interacting with the current user and updating
 * DAS users.
 *
 **/
declare class usersService {
    /**
     * Gets the current user that was loaded with loadCurrentUser.
     *
     * Returns the current user object from CH.
     **/
    static getCurrentUser(): Object;

    /**
     * @name fortisoar.usersService#loadCurrentUser
     * @methodOf fortisoar.usersService
     *
     * Loads the current user object from CH and store it in local storage. If it
     * already has been loaded, it retrieves it from local storage cache.
     *
     * Returns the promise that will resolve with the current user object.
     * 
     * @param {Boolean=} forceRefresh If true then data fetched from server, else uses browsers local storage data.
     **/
    static loadCurrentUser(forceRefresh?: Boolean): Promise<any>;

    /**
     * Gets the avatar img src with base64 encoded image data for the given IRI.
     * If the avatar IRI is empty, it will return the default avatar URI.
     *
     * Returns promise that resolves with the image src, which is
     * either the default image or base 64 encoded image
     * 
     * @param {String} avatarIri The CH IRI that it will retrieve the image from
     * @param {Boolean=} playbookUser To set default avatar as per user
     **/
    static getAvatar(avatarIri: String, playbookUser?: Boolean): Promise<any>;

    /**
     * Gets the current user avatar from local storage if it exists, otherwise
     * loads it using getAvatar and saves it in local storage.
     *
     * Returns promise that resolves with the image src, which is
     * either the default image or base 64 encoded image
     **/
    static getCurrentAvatar(): Promise<any>;

    /**
     * Gets the user details based on the IRI.
     * Returns promise that resolves with the user details.
     *
     * @param {String} userIri The CH IRI that it will retrieve the user details from.
     **/
    static getUserByIri(userIri: String): Promise<any>;
}

/**
 * @name fortisoar.ViewTemplateService
 * @module fortisoar
 *
 * The `ViewTemplateService` loads and saves view templates.
 *
 **/
declare class ViewTemplateService {
    /**
     * Checks local storage for the requested page and returns it.
     * If it is not in storage, it falls back to checking whether another promise
     * to get it has already been set. If one has not been set, then an API
     * request is made to retrieve it and store it in local storage.
     * Returns a promise that will be resolved with the page once retrieved.
     *
     * @param {String} page The name of the page to be retrieved.
     * @param {Boolean=} noRedirect To handle redirection.
     * @param {Object=} queryParams Query parameter details to set record limit, page number or search query. 
     * @param {String=} templateUUID UUID of the template.
     * @param {Boolean=} toExport To set export flag in query.
     **/
    static get(page: String, noRedirect?: Boolean, queryParams?: Object, templateUUID?: String, toExport?: Boolean): Promise<any>;

    /**
     * Changes the structure of a row/column based view template widget.
     * Used when switching the number of columns or layout of the columns.
     * Returns the new set of columns after the restructuring
     *
     * @param {any} structure The structure to switch to. Could be numeric number
     * of columns, or array of bootstrap column classes (['col-md-2', 'col-md-10'])
     * @param {Array} currentColumns Array of the current columns that will be restructured
     * @param {String} objectName The level name below columns (widgets or fields)
     **/
    static changeStructure(structure: any, currentColumns: Array<Object>, objectName: String): Array<Object>;

    /**
     * Fills array with fields used in conditional filter.
     * Used while handling widget and tabs visibility settings in View Panels.
     * 
     * @param {Object} filter Conditional filter object.
     * @param {Array} conditionalFields Array of conditional fields filled recursively.
     **/
    static populateConditionalFields(filter: Object, conditionalFields: Array<String>): void;

    /**
     * API call to fetch records for the entity as per provided visibility filter.
     * Used while handling widget and tabs visibility settings on Dashboard, Reports and View Panels.
     * Returns a promise that will be resolved with the conditional visibility filtered data.
     * 
     * @param {Object} entity An object of entity.
     * @param {Object} query Query parameters to fetch the filtered data.
     * @param {Boolean=} isRelationshipFieldPresent If present, then record specific data is fetched.
     **/
    static getConditionalVisibilityFilteredData(entity: Object, query: Object, isRelationshipFieldPresent?: Boolean): Promise<any>;
}

/**
 * @name fortisoar.websocketService
 * @module fortisoar
 *
 * The `websocketService` handles the websocket functionality to update the UI and backend with 
 * live data
 *
 **/
declare class websocketService {
    /**
     * The method gets the web socket session id.
     * 
     * Returns Web socket session id.
     * 
     **/
    static getWebsocketSessionId(): String;

    /**
     * The method generates a fresh web socket session id.
     * 
     * Returns generated Web socket session id.
     **/
    static generateWebsocketSessionId(): String;

    /**
     * This method opens the connection through the socket URL
     * Returns a promise object that on success returns the websocket connection.
     * 
     * @param {Boolean=} reconnect If true, then broadcasts reconnect message.
     * 
     **/
    static openSocket(reconnect?: Boolean): Promise<any>;

    /**
     * This method gives a warning to manually refresh and reconnect the web socket due to inactivity.
     *
     **/
    static reconnect(): void;

    /**
     * This method helps to receive data on a successful subscription.
     * Returns a promise object that on success will return the subscription.
     * 
     * @param {String} topic title name given to subscribe the web socket.
     * @param {Function} callback function to be called when web socket is subscribed.
     * 
     **/
    static subscribe(topic: String,callback: Function): Promise<any>;

    /**
     * This method is used to unsubscribe the web socket for the provided subscription.
     * 
     * @param {any} subscription data passed to indicate the web socket was subscribed
     *
     **/
    static unsubscribe(subscription: any): void;

    /**
     * This method is used to disconnect the web socket.
     * 
     * @param {Boolean} logout boolean true passed to disconnect the web socket.
     *
     **/
    static disconnect(logout: Boolean): void;

    /**
     * This method is used to get the header token
     * 
     * Returns an authorization token.
     **/
    static getHeader(): String;
}

/**
 * @name fortisoar.widgetTemplateService
 * @module fortisoar
 *
 * The `widgetTemplateService` loads widget definition.
 *
 **/
declare class widgetTemplateService {
    /**
     * This function is used to create widget data in generic format which will be 
     * used to render widget in different layout like popup, half widget or full widget.
     * 
     * Returns widget definition object with all required widget properties.
     * Widget definition object properties: title, name, size, version, section, description,
     * icon, templateUrl, controller, controllerPath, editTemplateUrl, editControllerPath,
     * editController, editable, visiblePages
     *
     * @param {Object} widget Widget detail object.
     **/
    static generateWidgetDefinition(widget: Object): Object;
}